/*
 *  PCM - Mu-Law conversion
 *  Copyright (c) 2000 by Abramo Bagnara <abramo@alsa-project.org>
 *
 *  Wrapper for linphone Codec class by Simon Morlat <simon.morlat@free.fr>
 */

#include <stdlib.h>
#include <string.h>
#include "g711ulaw.h"

typedef short int int16_t;

/* some utilities private functions*/
static inline int val_seg(int val);
static inline int ulaw_to_s16(unsigned char u_val);
static inline unsigned char s16_to_ulaw(int pcm_val);

/* functions not called directly*/
void g711ulaw_encode(Codec *obj,char *buf, char *ulaw);
void g711ulaw_decode(Codec *obj,char *ulaw, char *buf);
void g711ulaw_getinfo(Codec *obj,struct codec_info *info);

/* the public codec_info structure*/
struct codec_info g711ulaw_info=
{
	"G711-U-law",
	G711ULAW_FRAME_SIZE,    /* size of the uncompressed frame*/
	G711ULAW_FRAME_SIZE/2,   /* size of compressed frame*/
	64000, /* bit rate*/
	{{
		8000
	}},               /* audio sampling freq*/
	g711ulaw_codec_new,   /* codec constructor*/
	0,				/* profile */
	CODEC_AUDIO,          /* type*/
	0,										/*usable, set later*/	
	1										/*usable for user, default value*/
};


static inline int val_seg(int val)
{
	int r = 0;
	val >>= 7;
	if (val & 0xf0) {
		val >>= 4;
		r += 4;
	}
	if (val & 0x0c) {
		val >>= 2;
		r += 2;
	}
	if (val & 0x02)
		r += 1;
	return r;
}

/*
 * s16_to_ulaw() - Convert a linear PCM value to u-law
 *
 * In order to simplify the encoding process, the original linear magnitude
 * is biased by adding 33 which shifts the encoding range from (0 - 8158) to
 * (33 - 8191). The result can be seen in the following encoding table:
 *
 *	Biased Linear Input Code	Compressed Code
 *	------------------------	---------------
 *	00000001wxyza			000wxyz
 *	0000001wxyzab			001wxyz
 *	000001wxyzabc			010wxyz
 *	00001wxyzabcd			011wxyz
 *	0001wxyzabcde			100wxyz
 *	001wxyzabcdef			101wxyz
 *	01wxyzabcdefg			110wxyz
 *	1wxyzabcdefgh			111wxyz
 *
 * Each biased linear code has a leading 1 which identifies the segment
 * number. The value of the segment number is equal to 7 minus the number
 * of leading 0's. The quantization interval is directly available as the
 * four bits wxyz.  * The trailing bits (a - h) are ignored.
 *
 * Ordinarily the complement of the resulting code word is used for
 * transmission, and so the code word is complemented before it is returned.
 *
 * For further information see John C. Bellamy's Digital Telephony, 1982,
 * John Wiley & Sons, pps 98-111 and 472-476.
 */

static inline unsigned char s16_to_ulaw(int pcm_val)	/* 2's complement (16-bit range) */
{
	int mask;
	int seg;
	unsigned char uval;

	if (pcm_val < 0) {
		pcm_val = 0x84 - pcm_val;
		mask = 0x7f;
	} else {
		pcm_val += 0x84;
		mask = 0xff;
	}
	if (pcm_val > 0x7fff)
		pcm_val = 0x7fff;

	/* Convert the scaled magnitude to segment number. */
	seg = val_seg(pcm_val);

	/*
	 * Combine the sign, segment, quantization bits;
	 * and complement the code word.
	 */
	uval = (seg << 4) | ((pcm_val >> (seg + 3)) & 0x0f);
	return uval ^ mask;
}

/*
 * ulaw_to_s16() - Convert a u-law value to 16-bit linear PCM
 *
 * First, a biased linear code is derived from the code word. An unbiased
 * output can then be obtained by subtracting 33 from the biased code.
 *
 * Note that this function expects to be passed the complement of the
 * original code word. This is in keeping with ISDN conventions.
 */
static inline int ulaw_to_s16(unsigned char u_val)
{
	int t;

	/* Complement to obtain normal u-law value. */
	u_val = ~u_val;

	/*
	 * Extract and bias the quantization bits. Then
	 * shift up by the segment number and subtract out the bias.
	 */
	t = ((u_val & 0x0f) << 3) + 0x84;
	t <<= (u_val & 0x70) >> 4;

	return ((u_val & 0x80) ? (0x84 - t) : (t - 0x84));
}

/*constructor*/

Codec * g711ulaw_codec_new()
{
	G711ulawCodec *codec;
	
	codec=malloc(sizeof(G711ulawCodec));
	codec->baseclass._encode=g711ulaw_encode;
	codec->baseclass._decode=g711ulaw_decode;
	codec->baseclass._getinfo=g711ulaw_getinfo;
	codec->baseclass._destroy=(void (*)(struct _Codec *obj))free;  /* nothing special to do*/
	return((Codec*)codec);
}

void g711ulaw_encode(Codec *obj,char *buf, char *ulaw)
{
	int i;
	unsigned char *ulawbuf=(unsigned char*)ulaw;
	int16_t *lin16=(int16_t*)buf;
	
	for(i=0;i<G711ULAW_FRAME_SIZE/2;i++)
	{
		ulawbuf[i]=s16_to_ulaw((int)(lin16[i]));
	}
	return ;
}

void g711ulaw_decode(Codec *obj,char *ulaw, char *buf)
{
	int i;
	unsigned char *ulawbuf=(unsigned char*)ulaw;
	int16_t *lin16=(int16_t*)buf;
	for(i=0;i<G711ULAW_FRAME_SIZE/2;i++)
	{
		lin16[i]=(int)ulaw_to_s16(ulawbuf[i]);
	}
	return ;
}

void g711ulaw_getinfo(Codec *obj,struct codec_info *info)
{
	memcpy(info,&g711ulaw_info,sizeof(struct codec_info));
};
