/*
  The mediastreamer library aims at providing modular media processing and I/O
	for linphone, but also for any telephony application.
  Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org
  										
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "msoss.h"
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <sys/soundcard.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <glib.h>

#define DEV_NAME_LEN 20
#define DEVICE_PREFIX "/dev/dsp"

#if GLIB_MAJOR_VERSION < 2
#define g_file_test(a,b) 1
#define G_FILE_TEST_EXISTS
#endif


MSOss oss_mgr;

static gint oss_scan_dev(MSOss *oss, gchar *devfile);

static void snd_card_init(SndCard *card)
{
	memset(card,0,sizeof(SndCard));
}

/* scans all possible /dev/... that may be oss device file and configure the MSOss obj according
to the results*/
gint ms_oss_autoconf(MSOss *obj)
{
	gint found;
	found=oss_scan_dev(obj,"/dev/dsp");  /*on linux it should work every time*/
	if (found) return 0;
	found=oss_scan_dev(obj,"/dev/audio"); /* for solaris, hpux ...*/
	if (found) return 0;
	g_warning("No oss devices found.");
	return -1;
}

static char * oss_get_card_name(char *devfile)
{
	int fd;
	fd=open(devfile,O_RDONLY|O_NONBLOCK);
	/* need to find a way to get a string to identify the card..*/
	return NULL;
}

/*scans one type of device file */
static gint oss_scan_dev(MSOss *obj, char *devfile)
{
	char *name,*mixname;
	gint i,len=strlen(devfile)+2;

	name=g_malloc(len);
	mixname=g_malloc(strlen("/dev/mixer")+2);
	for (i=0;i<MSOSS_MAX_DEVICES;i++)
	{
		sprintf(name,"%s%i",devfile,i);
		if (g_file_test(name,G_FILE_TEST_EXISTS))
		{
			ms_oss_set_dev_name(obj,i,name);
			obj->wcards++;
			g_message("Detected %s - ",name);
		}
		else obj->card[i].dev_name=NULL;
		sprintf(mixname,"%s%i","/dev/mixer",i);
		if (g_file_test(mixname,G_FILE_TEST_EXISTS))
		{
			ms_oss_set_mixdev_name(obj,i,mixname);
		}
		else
		{
		   obj->card[i].mixdev_name=NULL;
		   g_warning("Unable to find mixer for id %i.",i);
		}
	}
	g_free(name);
	return obj->wcards;
}

void ms_oss_init(MSOss *klass)
{
	int i;

	for(i=0;i<MSOSS_MAX_DEVICES;i++)
	{
		snd_card_init(&klass->card[i]);
	}
	/* find devices */
	ms_oss_autoconf(klass);
}

/* open a sound device for duplex */
int ms_oss_opendev(MSOss *obj,int devid, int bits, int stereo, int rate, int blocksize)
{
	int fd,p;

	SndCard *card=&obj->card[devid];
	gint min_size,cond,i;

	g_return_val_if_fail(devid<MSOSS_MAX_DEVICES,-EINVAL);
	g_return_val_if_fail(card->fd==0,-EBUSY);
	g_return_val_if_fail(card->dev_name!=NULL,-EFAULT);

	/* open the device */
	fd=open(card->dev_name,O_RDWR);

	g_return_val_if_fail(fd>0,-errno);

	ioctl(fd, SNDCTL_DSP_RESET, 0);

	p =  bits;  /* 16 bits */
	ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &p);

	p =  stereo;  /* number of channels */
	ioctl(fd, SNDCTL_DSP_CHANNELS, &p);

	p =  rate;  /* rate in khz*/
	ioctl(fd, SNDCTL_DSP_SPEED, &p);

	ioctl(fd, SNDCTL_DSP_GETBLKSIZE, &min_size);

	/* try to subdivide BLKSIZE to reach blocksize if necessary */
	if (min_size>blocksize)
  {
		cond=1;
    	p=min_size/blocksize;
    	while(cond)
    	{
			i=ioctl(fd, SNDCTL_DSP_SUBDIVIDE, &p);
			printf("SUB_DIVIDE said error=%i,errno=%i\n",i,errno);
     	if ((i==0) || (p==1)) cond=0;
     	else p=p/2;
     }
	}
	ioctl(fd, SNDCTL_DSP_GETBLKSIZE, &min_size);
	if (min_size>blocksize)
	{
		g_warning("dsp block size set to %i.",min_size);
	}
	g_message("dsp blocksize is %i.",min_size);
	card->fd=fd;
	card->bits=bits;
	card->stereo=stereo;
	card->rate=rate;
	card->blocksize=min_size;
	return 0;
}
/* close a sound device */
void ms_oss_closedev(MSOss *obj,int devid)
{
	SndCard *card=&obj->card[devid];
	g_return_if_fail(devid<MSOSS_MAX_DEVICES);
	if (card->fd!=0) close(card->fd);
}

/* set the path to device id */
void ms_oss_set_dev_name(MSOss *klass,int devid, char *path)
{
	SndCard *card;
	g_return_if_fail(devid<MSOSS_MAX_DEVICES);
	card=&klass->card[devid];
	if (card->dev_name!=NULL) g_free(card->dev_name);
	card->dev_name=g_strdup(path);;
}

/* set the path to mixer device id */
void ms_oss_set_mixdev_name(MSOss *klass,int devid, char *path)
{
	SndCard *card;
	g_return_if_fail(devid<MSOSS_MAX_DEVICES);
	card=&klass->card[devid];
	if (card->mixdev_name!=NULL) g_free(card->mixdev_name);
	card->mixdev_name=g_strdup(path);;
}

/*set recording level */
void ms_oss_set_rlevel(MSOss *obj, gint devid, gint a)
{
   SndCard *card;
   int p,mix_fd;
	g_return_if_fail(devid<MSOSS_MAX_DEVICES);
	card=&obj->card[devid];
	g_return_if_fail(card->mixdev_name!=NULL);

	p=(((int)a)<<8 | (int)a);
	mix_fd = open(card->mixdev_name, O_WRONLY);
	ioctl(mix_fd,MIXER_WRITE(SOUND_MIXER_RECLEV), &p);
	close(mix_fd);
}

/*set playback level */
void ms_oss_set_plevel(MSOss *obj, gint devid, gint a)
{
   SndCard *card;
   int p,mix_fd;
	g_return_if_fail(devid<MSOSS_MAX_DEVICES);
	card=&obj->card[devid];
	g_return_if_fail(card->mixdev_name!=NULL);

	p=(((int)a)<<8 | (int)a);
	mix_fd = open(card->mixdev_name, O_WRONLY);
	ioctl(mix_fd,MIXER_WRITE(SOUND_MIXER_PCM), &p);
	close(mix_fd);
}

