/*
  The oRTP LinPhone RTP library intends to provide basics for a RTP stack.
  Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "rtpsession.h"
#include "rtpmod.h"

#ifdef TARGET_IS_HPUXKERNEL

#else
#include <fcntl.h>
#include <sys/types.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>
#endif

static gint normal_ts_inc[10]={
/*0=Alaw*/	160,
/*1=?   */    0,
/*2=?   */    0,
/*3=GSM */  160,
/*4=LPC */  180,
/*5=?   */    0,
/*6=?   */    0,
/*7=?   */    0,
/*8=Mulaw*/ 160,
/*9=?   */    0
};

#ifdef HAVE_GLIB
void dummy_log(const gchar *log_domain,
                                             GLogLevelFlags log_level,
                                             const gchar *message,
                                             gpointer user_data)
{
	return;
}
#endif


void ortp_init()
{
#ifdef HAVE_GLIB
	g_log_set_handler ("oRTP", G_LOG_LEVEL_MESSAGE, dummy_log, NULL);
#endif
}

#ifdef HAVE_GLIB
void ortp_log( const gchar *log_domain,
                                             GLogLevelFlags log_level,
                                             const gchar *message,
                                             gpointer user_data)
{
	fprintf((FILE*) user_data,"%s: %s\n",log_domain,message);
}
#endif

#ifdef HAVE_GLIB
#ifndef TARGET_IS_HPUXKERNEL
void ortp_set_debug_file(FILE *file)
{
	if (file!=NULL)
		g_log_set_handler ("oRTP", G_LOG_LEVEL_MESSAGE, ortp_log, (gpointer)file);
	else g_log_set_handler ("oRTP", G_LOG_LEVEL_MESSAGE, dummy_log, NULL);
}
#endif
#endif


void rtp_session_init(RtpSession *session,gint type)
{
	memset(session,0,sizeof(RtpSession));
	session->rtp.max_rq_size=RTP_MAX_RQ_SIZE;
	session->rtp.jitt_comp_time=RTP_DEFAULT_JITTER;
	rtp_session_set_flag(session,type|RTP_SESSION_IS_STARTING);
#ifndef TARGET_IS_HPUXKERNEL
	session->rtp.rq=&session->rtp._rq;
	session->rtp.wq=&session->rtp._wq;
#endif
}

RtpSession *rtp_session_new(gint type)
{
	RtpSession *session=g_malloc(sizeof(RtpSession));
	rtp_session_init(session, type);
	return session;
}

#ifdef TARGET_IS_HPUXKERNEL
gint rtp_session_set_local_addr(RtpSession *session, gchar *addr,gint port)
{
	return EOPNOTSUPP;
}
#else
gint rtp_session_set_local_addr(RtpSession *session, gchar *addr,gint port)
{
	gint err;
	
	session->rtp.loc_addr.sin_family=AF_INET;
	
	err=inet_aton(addr,&session->rtp.loc_addr.sin_addr);
	if (err<0)
	{
		g_warning("Error in socket address:%s.",strerror(errno));
		return err;
	}
	session->rtp.loc_addr.sin_port=htons(port);
	
	session->rtp.socket=socket(PF_INET,SOCK_DGRAM,0);
	g_return_val_if_fail(session->rtp.socket>0,-1);
	/* set non blocking mode */
	fcntl(session->rtp.socket,F_SETFL,O_NONBLOCK);	
	err=bind(session->rtp.socket,(struct sockaddr*)&session->rtp.loc_addr,sizeof(struct sockaddr_in));
	if (err!=0)
	{
		g_warning("Fail to bind rtp socket: %s.",strerror(errno));
		close(session->rtp.socket);
		return -1;
	}
	
	memcpy(&session->rtcp.loc_addr,&session->rtp.loc_addr,sizeof(struct sockaddr_in));
	session->rtcp.loc_addr.sin_port=htons(port+1);
	session->rtcp.socket=socket(PF_INET,SOCK_DGRAM,0);
	g_return_val_if_fail(session->rtcp.socket>0,-1);
	err=bind(session->rtcp.socket,(struct sockaddr*)&session->rtcp.loc_addr,sizeof(struct sockaddr_in));
	if (err!=0)
	{
		g_warning("Fail to bind rtcp socket: %s.",strerror(errno));
		close(session->rtp.socket);
		close(session->rtcp.socket);
		return -1;
	}
	FD_ZERO(&session->scanfd);
	FD_SET(session->rtp.socket,&session->scanfd);
	FD_SET(session->rtcp.socket,&session->scanfd);
	if (session->rtcp.socket> session->rtp.socket) session->highest_fd=session->rtcp.socket+1;
	else session->highest_fd=session->rtp.socket+1;
	return 0;
}
#endif

#ifdef TARGET_IS_HPUXKERNEL
gint rtp_session_set_remote_addr(RtpSession *session, gchar *addr,gint port)
{
	return EOPNOTSUPP;
}
#else
gint rtp_session_set_remote_addr(RtpSession *session, gchar *addr,gint port)
{
	gint err;
	
	if (session->rtp.socket==0)
	{
		/* the session has not its socket bound, do it*/
		g_message("Setting random local addresses.");
		rtp_session_set_local_addr(session,"0.0.0.0",((rand()+5000) & 0xffff));
	}
	session->rtp.rem_addr.sin_family=AF_INET;
	
	err=inet_aton(addr,&session->rtp.rem_addr.sin_addr);
	if (err<0)
	{
		g_warning("Error in socket address:%s.",strerror(errno));
		return err;
	}
	session->rtp.rem_addr.sin_port=htons(port);
	
	memcpy(&session->rtcp.rem_addr,&session->rtp.rem_addr,sizeof(struct sockaddr_in));
	session->rtcp.rem_addr.sin_port=htons(port+1);
	
	if (session->flags & RTP_SESSION_DUPLEX)
	{
  	err=connect(session->rtp.socket,(struct sockaddr*)&session->rtp.rem_addr,sizeof (struct sockaddr_in));
  	if (err!=0)
  	{
  		g_message("Can't connect rtp socket: %s.",strerror(errno));
  		return err;
  	}
  	err=connect(session->rtcp.socket,(struct sockaddr*)&session->rtcp.rem_addr,sizeof (struct sockaddr_in));
  	if (err!=0)
  	{
  		g_message("Can't connect rtp socket: %s.",strerror(errno));
  		return err;
  	}
  }
	return 0;
}	
#endif

static void rtp_header_write(rtp_header_t *rtp,guint32 ts,guint16 seq, guint8 paytype, guint32 ssrc)
{
	rtp->version=2;
	rtp->padbit=0;
	rtp->extbit=0;
	rtp->cc=0;
	rtp->paytype=paytype;
	rtp->timestamp=htonl(ts);
	rtp->seq_number=htons(seq);
	rtp->ssrc=htonl(ssrc);
}

#ifdef TARGET_IS_HPUXKERNEL
static gint rtp_stack_send(RtpSession *session)
{
	mblk_t *mp=getq(session->rtp.wq);
	/* REVISIT, it may not be very efficient */
	streams_put(putnext,session->rtp.wq,mp,session->rtp.wq);
	return 0;
}
#else
static gint rtp_stack_send(RtpSession *session)
{
	mblk_t *m;
	gint error;
	m=getq(session->rtp.wq);
	
	error=sendto(session->rtp.socket,m->b_rptr,(m->b_wptr-m->b_rptr),0,
								(struct sockaddr*)&session->rtp.rem_addr,sizeof(struct sockaddr_in));
	if (error<0) g_warning("Error sending rtp packet: %s.",strerror(errno));
	return error;
}							

#endif

/**
 *rtp_session_set_jitter_compensation:
 *@session: a RtpSession
 *@milisec: the time interval in milisec to be jitter compensed.
 *
 * Sets the time interval for which packet are buffered instead of being delivered to the 
 * application.
 **/
int rtp_session_set_jitter_compensation(RtpSession *session, int milisec)
{
	/* REVISIT: need to take in account the payload description*/
	session->rtp.jitt_comp=milisec;
	session->rtp.jitt_comp_time=milisec*8;
}

gint rtp_session_send_with_ts(RtpSession *session, gchar *buffer, gint len, guint32 time)
{
	mblk_t *m;
	rtp_header_t *rtp;
	guint32 timestamp;
	gint error=0;
	
	timestamp=session->rtp.snd_ts_offset+time;
	if (RTP_TIMESTAMP_IS_STRICTLY_NEWER_THAN(session->rtp.snd_ts,timestamp))
	{
		g_warning("Bad timestamp !");
		return -1;
	}
	
	/* allocate a mblk_t, set the haeder. Perhaps if len>MTU, we should allocate a new
	mlkt_t to split the packet FIXME*/
	m=allocb(len+RTP_FIXED_HEADER_SIZE,BPRI_MED);
	rtp=(rtp_header_t*) m->b_wptr;
	rtp_header_write(rtp,timestamp,session->rtp.snd_seq,session->profile,session->ssrc);
	m->b_wptr+= RTP_FIXED_HEADER_SIZE;
	memcpy(m->b_wptr,buffer,len);
	m->b_wptr+=len;
	session->rtp.snd_seq++;
#ifdef TARGET_IS_HPUXKERNEL
	/* send directly the message through the stream */
	streams_put(putnext,session->rtp.wq,m,(void*)session->rtp.wq);
#else
	putq(session->rtp.wq,m);
	if (session->flags & RTP_SESSION_NO_SYNC)
	{
		error=rtp_stack_send(session);
	}
	else
	{
		g_warning("Synchronised mode not availlable for the moment !");
	}
#endif
	return error;
}


#ifdef TARGET_IS_HPUXKERNEL
static gint rtp_stack_recv(RtpSession *session)
{
	return EOPNOTSUPP;
}
#else
static gint rtp_stack_recv(RtpSession *session)
{
	gint error=1;
	struct sockaddr_in remaddr;
	int addrlen=sizeof( struct sockaddr_in);
	char *p;
	fd_set tmpfdset;
	mblk_t *mp=allocb(UDP_MAX_SIZE,0);
	
	FD_ZERO(&tmpfdset);
	if (session->flags & RTP_SESSION_IS_BLOCKING)
	{
		while (error>0 && (!FD_ISSET(session->rtp.socket,&tmpfdset)))
		{
			memcpy(&tmpfdset,&session->scanfd,sizeof(fd_set));
			error=select(session->highest_fd,&tmpfdset,NULL,NULL,session->rtp.timeout);
			g_message("select returned %i",error);
		}
	}
	
	error=recvfrom(session->rtp.socket,mp->b_wptr,UDP_MAX_SIZE,0,(struct sockaddr*)&remaddr,&addrlen);
	if (error>0)
	{
		/* resize the memory allocated to fit the udp message */
		
		p=g_realloc(mp->b_wptr,error);
		if (p!=mp->b_wptr) g_message("The recv area has moved during reallocation.");
		mp->b_datap->db_base=mp->b_rptr=mp->b_wptr=p;
		mp->b_wptr+=error;
	  mp->b_datap->db_lim=mp->b_wptr;
	  /* then put the new message on queue */
		rtp_parse(session,mp);
	}
	else
	{
		g_warning("Error receiving udp packet: read=%i.",error);
		if (error<0) g_warning("%s",strerror(errno));
		freeb(mp);
	}
	return error;
}
#endif

mblk_t * rtp_session_recvm_with_ts(RtpSession *session, guint32 timestamp)
{
	mblk_t *mp;
	rtp_header_t *rtp;
	guint32 ts;
	
#ifdef TARGET_IS_HPUX_KERNEL
	/* nothing to do: packets are enqueued on interrupt !*/
#else
	if (session->flags & RTP_SESSION_NO_SYNC)
	{
		/* recv until socket is empty*/
		if (!(session->flags & RTP_SESSION_IS_BLOCKING))
			while (rtp_stack_recv(session)>0);
		else
		{
			rtp_stack_recv(session);
		}
	}
#endif
	/* then now try to return a packet, if possible */
	/* first condition: if the session is starting, don't return anything
	until the queue size reaches jitt_comp */
	if (session->flags & RTP_SESSION_IS_STARTING)
	{
		rtp_header_t *oldest,*newest;
		queue_t *q=session->rtp.rq;
		if (q->q_last==NULL){
			g_message("Queue is empty.");
			return NULL;
		}
		oldest=(rtp_header_t*) q->q_first->b_rptr;
		newest=(rtp_header_t*) q->q_last->b_rptr;
		if ( (guint32)(newest->timestamp-oldest->timestamp) < session->rtp.jitt_comp_time)
		{
			g_message("Not enough packet bufferised.");
			return NULL;
		}
		/* if enough packet queued continue but delete the starting flag */
		rtp_session_unset_flag(session,RTP_SESSION_IS_STARTING);
		
		/* and also remember the timestamp offset between the session timestamp (random)
		and the user timestamp, that very often starts at zero */
		mp=getq(session->rtp.rq);
		rtp=(rtp_header_t*) mp->b_rptr;
		session->rtp.rcv_ts_offset=(guint32)(rtp->timestamp-timestamp);
		g_message("Returning FIRST mp with ts=%i",rtp->timestamp);
		return mp;
	}
	/* else this the normal case */
	/*calculate the real timestamp */
	ts=timestamp+session->rtp.rcv_ts_offset;
	mp=rtp_getq(session->rtp.rq,ts);
	if (mp==NULL)
	{
		g_message("No mp for timestamp queried");
		return NULL;
	}
	rtp=(rtp_header_t*)mp->b_rptr;
	g_message("Returning mp with ts=%i",rtp->timestamp);
	/* perhaps we can now make some checks to see if a resynchronization is needed */
	/* TODO*/
	return mp;
}


gint rtp_session_recv_with_ts(RtpSession *session, gchar *buffer, gint len, guint32 time)
{
	mblk_t *mp,*mprev,*mnext;
	gint rlen=len;
	
	mp=rtp_session_recvm_with_ts(session,time);
	if (mp==NULL)
	{		
		memset(buffer,0,len); /* ?? */
		return 0;
	}
	/* copy the payload in the user buffer */
	mnext=mp->b_cont;
	mprev=mp;
	while(mnext!=NULL)
	{
		gint msglen= mnext->b_wptr-mnext->b_rptr;
		if (msglen<=rlen)   {
			memcpy(buffer,mnext->b_rptr,msglen);
			/* go to next mblk_t */
			mprev->b_cont=mnext->b_cont;
			mprev=mnext;
			mnext=mnext->b_cont;
			mprev->b_cont=NULL;
			freeb(mprev);
			/* the current mblk_t (mnext) is unlinked from the prev. The prev is linked to next->b_cont->b_cont*/
		}
		else
		{
			g_warning("Rtp packet does not fit in supplied buffer !");
			memcpy(buffer,mnext->b_rptr,rlen);
			mnext->b_rptr+=rlen;
			/* re - enqueue the packet for next call */
			insq(session->rtp.rq,NULL,mp);
			return(len);
		}
		buffer+=msglen;
		rlen-=msglen;
	}
	/* free mp*/
	freemsg(mp);
	return (len-rlen);	
}


int rtp_session_recv(RtpSession *session, gchar *buffer, gint len)
{
	session->rtp.rcv_ts+=normal_ts_inc[session->profile];
	return(rtp_session_recv_with_ts(session,buffer,len,session->rtp.rcv_ts));
}

int rtp_session_send(RtpSession *session, gchar *buffer, gint len)
{
	session->rtp.snd_ts+=normal_ts_inc[session->profile];
	return(rtp_session_send_with_ts(session,buffer,len,session->rtp.snd_ts));
}

#ifdef TARGET_IS_HPUXKERNEL
void rtp_session_set_timeout(RtpSession *session, guint milisec)
{
	return;
}
#else
void rtp_session_set_timeout(RtpSession *session, guint milisec)
{
	if (milisec==0)
	{
		session->rtp.timeout=NULL;
		return;
	}
	session->rtp._timeout.tv_sec=milisec/1000;
	session->rtp._timeout.tv_usec=(milisec % 1000)*1000000;
	session->rtp.timeout=&session->rtp._timeout;
}
#endif

void rtp_session_uninit(RtpSession *session)
{
	/*flush all queues */
	flushq(session->rtp.rq,FLUSHALL);
	flushq(session->rtp.wq,FLUSHALL);
#ifndef TARGET_IS_HPUXKERNEL
	/* close sockets */
	close(session->rtp.socket);
	close(session->rtcp.socket);
#endif
}

void rtp_session_destroy(RtpSession *session)
{
	rtp_session_uninit(session);
	g_free(session);
}
