/*
  The osipua library is a library based on oSIP that implements CallLeg and User Agent
  level.
  Copyright (C) 2001  Simon MORLAT simon.morlat@free.fr
  											Aymeric MOIZARD jack@atosc.org
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <osip/osip.h>
#include <osip/port.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <errno.h>
#include "osipua.h"
#include "udp.h"
#include "osipmanager.h"
#include "resolver.h"


static void set_all_callbacks(osip_t *cf);

#define unblock_udp_thread(manager) write(manager->udp_control_fd,(void*)manager,1)

OsipManager* osip_manager_new()
{
	OsipManager *manager;
	int control_fds[2];
	int err;
	
	manager=smalloc(sizeof(OsipManager));
	memset(manager,0,sizeof(OsipManager));
	if (-1==osip_global_init())
	{
		osip_trace(TRACE_LEVEL0,("error: Failed to init mutex.\n"));
		return NULL; /* mutex is not initialised properly */
	}
	osip_init(&manager->config);
	
#if defined(THREAD_PTH)
	pth_init();
#endif	
	set_all_callbacks(manager->config);
	/* create the control file descriptor (used to unblock the udp daemon sometimes*/
	err=pipe(control_fds);
	if (err!=0)
    {
      perror("Error creating pipe");
      exit(1);
    }
	FD_SET(control_fds[0],&manager->udpfdset);
	manager->udp_unblock_fd=control_fds[0];
  	manager->udp_control_fd=control_fds[1]; /* the file descriptor where to write something to control the daemon*/
	manager->max_udpfd=control_fds[0];
	manager->recv_tout.tv_sec=0;
	manager->recv_tout.tv_usec=100000;  /*100 ms lets good precision for retransmission that should occur every 500 ms*/
	/* allocate the memory to receive packets*/
	manager->udp_buf = (char *)smalloc(SIP_MESSAGE_MAX_LENGTH*sizeof(char)+1);
	manager->resolv_fifo=smalloc(sizeof(fifo_t));
	fifo_init(manager->resolv_fifo);
	return manager;
}

void osip_manager_destroy(OsipManager *manager)
{
	int i;
	
	osip_manager_stop_udp_daemon(manager);
	sfree(manager->udp_buf);
	/* close all file descripors */
	for (i=0;i<OSIP_MAX_UDP_PORTS;i++)
	{
		if (manager->udpfds[i]!=0) close( manager->udpfds[i]);
	}
	/* free the osip_t structure : to do*/
	sfree(manager->config);
	fifo_free(manager->resolv_fifo);
	sfree(manager->resolv_fifo);
	sfree(manager);
}
	

int osip_manager_add_udpport(OsipManager *manager, int port)
{
	int i,newfd,err;
	struct sockaddr_in laddr;
	int option=1;
	
	for(i=0;i<OSIP_MAX_UDP_PORTS;i++)
	{
		if (port==manager->udpports[i])
		{
			osip_trace(TRACE_LEVEL3,("info: port already listened\n"));
			return -EALREADY;
		}
	}
	if ( port!=manager->send_port)
	{ 	 /* create a new file descriptor*/
       newfd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

    	laddr.sin_addr.s_addr = htons(INADDR_ANY);
    	laddr.sin_port = htons((short)port);
    	laddr.sin_family = AF_INET;
    	if (bind(newfd, (struct sockaddr *)&laddr, sizeof(laddr)) < 0)
    	{
    		osip_trace(TRACE_LEVEL0,("error: Failed to bind socket !\n"));
    		close(newfd);
    		return -errno;
    	}
    	err=setsockopt(newfd, SOL_SOCKET , SO_REUSEADDR, (void*)&option, sizeof(option));
    	if (err!=0) osip_trace(TRACE_LEVEL3,("warning: port cannot be reused.\n"));
	}
	else
	{
		/* if the socket used to send data has the same port as the new port to scan, then use it*/
    	newfd=manager->send_sock;
	}
	for (i=0;i<OSIP_MAX_UDP_PORTS;i++)
	{
		if (  manager->udpports[i]==0)
		{
			/* this is a free place to store the port number and file descriptor*/
			manager->udpports[i]=port;
			manager->udpfds[i]=newfd;
			FD_SET(newfd,&manager->udpfdset);
			if (newfd>manager->max_udpfd) manager->max_udpfd=newfd;
			/* unblock the udp daemon thread*/
			unblock_udp_thread(manager);
			return 0;
		}
	}
	/* no more port availlable for listening*/
	close(newfd);
	osip_trace(TRACE_LEVEL3,("info: The table of scanned file descriptor is full.\n"));
	return -1;
}
	
int osip_manager_remove_udpport(OsipManager *manager, int port)
{
   int i;
	
	for(i=0;i<OSIP_MAX_UDP_PORTS;i++)
	{
		if (port==manager->udpports[i])
		{
			manager->udpports[i]=0;
			FD_CLR(manager->udpfds[i],&manager->udpfdset);
			/* unblock the udp dameon*/
			unblock_udp_thread(manager);
			if (port!=manager->send_port) close(manager->udpfds[i]);
			manager->udpfds[i]=0;
			return 0;
		}
	}
	osip_trace(TRACE_LEVEL3,("error: Could not close Udp Port.\n"));
	return -ENOENT;
}


void osip_manager_stop_udp_daemon(OsipManager *manager)
{
	if (manager->udp_run_cond)
	{
		manager->udp_run_cond=0;
		unblock_udp_thread(manager);
		sthread_join(&manager->udp_thread);
	}
}

void osip_manager_start_udp_daemon(OsipManager *manager)
{
	/* start a daemon*/
	if (manager->udp_run_cond==1) return;
	manager->udp_run_cond=1;
	/* create daemon thread*/
	sthread_create(0,&manager->udp_thread,sipd_thread,(void*)manager);
	return; /* ok */
}

void osip_manager_start_resolver(OsipManager *manager)
{
	/* start the resolver thread*/
	if (manager->resolv_run_cond==1) return;
	manager->resolv_run_cond=1;
	/* create daemon thread*/
	sthread_create(0,&manager->resolv_thread,resolver_thread,(void*)manager);
	return; /* ok */
}

void osip_manager_stop_resolver(OsipManager *manager)
{
	if (manager->resolv_run_cond)
	{
		manager->resolv_run_cond=0;
		sthread_join(&manager->resolv_thread);
	}
}

static void set_all_callbacks(osip_t *cf)
{
  //osip_setcb_rcvresp_retransmission(cf,&cb_rcvresp_retransmission);
  //osip_setcb_sndreq_retransmission(cf,&cb_sndreq_retransmission);
  //osip_setcb_sndresp_retransmission(cf,&cb_sndresp_retransmission);
  //osip_setcb_rcvreq_retransmission(cf,&cb_rcvreq_retransmission);

  osip_setcb_killtransaction(cf,&cb_killtransaction);
  osip_setcb_endoftransaction(cf,&cb_endoftransaction);

  osip_setcb_connection_refused(cf,&cb_connection_refused);
  osip_setcb_network_error(cf,&cb_network_error);

  osip_setcb_rcv1xx(cf,&cb_rcv1xx);
  osip_setcb_rcv2xx(cf,&cb_rcv2xx);
  osip_setcb_rcv3xx(cf,&cb_rcv3xx);
  osip_setcb_rcv4xx(cf,&cb_rcv4xx);
  osip_setcb_rcv5xx(cf,&cb_rcv5xx);
  osip_setcb_rcv6xx(cf,&cb_rcv6xx);

  //osip_setcb_sndinvite  (cf,&cb_sndinvite);
  //osip_setcb_sndack     (cf,&cb_sndack);
  //osip_setcb_sndbye     (cf,&cb_sndbye);
  //osip_setcb_sndcancel  (cf,&cb_sndcancel);
  //osip_setcb_sndinfo    (cf,&cb_sndinfo);
  //osip_setcb_sndoptions (cf,&cb_sndoptions);
  //osip_setcb_sndregister(cf,&cb_sndregister);
  //osip_setcb_sndprack   (cf,&cb_sndprack);
  //osip_setcb_sndunkrequest(cf,&cb_sndunkrequest);

  //osip_setcb_snd1xx(cf,&cb_snd1xx);
  //osip_setcb_snd2xx(cf,&cb_snd2xx);
  //osip_setcb_snd3xx(cf,&cb_snd3xx);
  //osip_setcb_snd4xx(cf,&cb_snd4xx);
  //osip_setcb_snd5xx(cf,&cb_snd5xx);
  //osip_setcb_snd6xx(cf,&cb_snd6xx);

  osip_setcb_rcvinvite  (cf,&cb_rcvinvite);
  osip_setcb_rcvack     (cf,&cb_rcvack);
  osip_setcb_rcvbye     (cf,&cb_rcvbye);
  osip_setcb_rcvcancel  (cf,&cb_rcvcancel);
  //osip_setcb_rcvinfo    (cf,&cb_rcvinfo);
  //osip_setcb_rcvoptions (cf,&cb_rcvoptions);
  //osip_setcb_rcvregister(cf,&cb_rcvregister);
  //osip_setcb_rcvprack   (cf,&cb_rcvprack);
  osip_setcb_rcvunkrequest(cf,&cb_rcvunkrequest);

}

int osip_manager_set_send_port(OsipManager *manager,int port)
{
	int newfd,err;
	struct sockaddr_in laddr;
	int option=1;
	
	if (manager->send_port==port) return 0;
	else if (manager->send_sock!=0) close(manager->send_sock);
	
	/* create a new file descriptor*/
   newfd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

	laddr.sin_addr.s_addr = htons(INADDR_ANY);
	laddr.sin_port = htons((short)port);
	laddr.sin_family = AF_INET;
	if (bind(newfd, (struct sockaddr *)&laddr, sizeof(laddr)) < 0)
	{
		osip_trace(TRACE_LEVEL0,("error: Could not to bind socket for sending messages.\n"));
		close(newfd);
		return -errno;
	}
	err=setsockopt(newfd, SOL_SOCKET , SO_REUSEADDR, (void*)&option, sizeof(option));
	if (err!=0) osip_trace(TRACE_LEVEL3,("warning: port cannot be reused.\n"));
	manager->send_sock=newfd;
	manager->send_port=port;
	return 0;
}

