/*
  The osipua library is a library based on oSIP that implements CallLeg and User Agent
  level.
  Copyright (C) 2001  Simon MORLAT simon.morlat@free.fr
  											Aymeric MOIZARD jack@atosc.org
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef OSIPUA_H
#define OSIPUA_H

#include <stdio.h>

#include <osip/osip.h>
#include <osip/smsgtypes.h>
#include <osip/sdp.h>
#include "mediadesc.h"

#define IP4_SIZE 20

/* used by the osip_trace macro*/
char *make_message(const char *fmt, ...);

#define osip_trace(loglevel,args)  do        \
{                       \
	char *__strmsg;  \
	__strmsg=make_message args ;    \
	TRACE(trace(__FILE__,__LINE__,(loglevel),stdout,__strmsg)); \
	sfree (__strmsg);        \
}while (0);

struct _OsipCallLeg
{
	from_t *from;
	to_t *to;
	call_id_t *callid;
	int cseq_number;
	int status;
	url_t *registrar;
	 /*
       * GRS
       * from and to are not enough if (for example) the far end proxy server <        * gets out of the loop and you end up communicating directly with the
       * end system.  Your from and to remain constant, but you need to
       * know who the "real" peer is for request URIs and to send the UDP
       * packets to the right place.
       *
       * this could be handled in one of two ways - the "peer" field could
       * stay null unless its valid, or we could always initialize it to
       * be the same as the to field.  Simon, I'll leave that up to you.
       */
	contact_t *peer;
	/* transactions associated with call-leg */
	list_t incoming_transactions;
	list_t outgoing_transactions;
	int last_tid;    /* the id of the last transaction*/
	
	struct _RegistrationCtxt *reg_context;  /* the registration context if a registration is occuring */
	unsigned int tag_set:1;
	unsigned int padding:31;
	
  /* all events for this transaction arrive here */
	//fifo_t *calllegff;
	struct _OsipUA *ua; /* the ua object that manages the call*/	

        /* BKO: list of routes to pass requests through, set after
           first INVITE */
        list_t route_list;
	
	/*sdp only parameters*/
	sdp_t *remote_sdp;
	int sdp_sessid;
	char* sdp_caddr;   /* the address in the c field*/
	/* list of used streams with their properties*/
	list_t *media_list;
};

typedef struct _OsipCallLeg OsipCallLeg;

/* the user agent class*/
struct _OsipUA
{
	osip_t *config; /* the config structure driving the stack*/
	struct _OsipManager *manager;  /* the manager object that manages the stack on which the ua is running */
	contact_t *contact;  /* the user agent real identity, like machin@217.0.0.128 */
	list_t *alias;      /* list of alias like machin@www-db.research.bell-labs.com, including address of records */
	char *fromtag;          /* the from tag put in all from fields of the requests */

	char ua_ip4addr[IP4_SIZE];
	int ua_port;
	list_t call_list;  /* the list of call-legs managed by the user agent*/
	int max_calls;  /* the number of call of ua will process. If the limit is exceeded, it will answer BUSY HERE*/
	url_t *registrar;   /* sip url of a registrar. This is also the address of the proxy if the OSIPUA_USE_PROXY flag is set */
	char *reg_passwd;  /* the password used for registration*/
	unsigned int flags;
	#define OSIP_UA_USE_PROXY (0x0001)
	struct _MediaDesc *maudio;     /* a MediaDescriptor for audio*/
	struct _MediaDesc *mvideo;     /* a MediaDescriptor for video*/
	/*signals*/
	int (*invite)(OsipCallLeg *,sip_t *, void*);   /*INVITE*/
	int (*invite_accepted)(OsipCallLeg *,sip_t*, void*);   /*INVITE_ACCEPTED*/    /* called to signal that media sessions can start*/
	int (*bye)(OsipCallLeg *,sip_t*, void*);  /*BYE*/
	int  (*faillure)(OsipCallLeg *,sip_t*, void*);  /*FAILLURE*/
	int  (*informative)(OsipCallLeg *,sip_t*, void*);  /*various informations*/
	FILE *dbg;

  /* presence management infos */
  int presence_mode;   /* 200-> accept calls
			  486-> answer automatic 486 Busy
			  600-> 600 Busy Evrywhere
			  480-> answar automatic 480 Temporirily unavailable
		          302-> Moved temporarily
		          380-> Alternative service */
  int presence_delay;     /* a value in seconds to ba added to Retry-After
			  in 486 Busy or 480 Temporirily unavailable */
  char *presence_contact_url;   /* contain a sip-url for 302, 301, 380.
			  TO BE DONE: Can also contain an email instead of url!
			  */

};

typedef struct _OsipUA OsipUA;

typedef int (*OsipUASignalHandler)(OsipCallLeg *,sip_t*, void*);

/* a structure to give parameters to the ua*/
typedef struct _ua_params
{
	char *contact;
	char *ua_ip4addr;
	int ua_port;
	int max_calls;                         /* the number of call of ua will process. If the limit is exceeded, it will answer BUSY HERE*/
	/* session capabilities */
	struct _MediaDesc *maudio;     /* a MediaDescriptor for audio*/    /* can be NULL*/
	struct _MediaDesc *mvideo;     /* a MediaDescriptor for video*/     /*can be NULL*/
	FILE *dbg;
} ua_params_t;

/* initialize the osipua library */
void osipua_init();

/* stop the threads runned by osipua */
void osipua_exit();

/* create new user agent with given parameters*/
OsipUA * osip_ua_new(ua_params_t *params);

/*set (or change) ua ip address*/
void osip_ua_set_ip4addr(OsipUA *ua,char *ip4addr);

/* change identity of an existent user agent*/
int osip_ua_set_contact(OsipUA *ua, char *contact);

/* add an alias identity to which the ua is supposed to answser */
int osip_ua_add_alias(OsipUA *ua, char *contact);

/* remove an alias*/
int osip_ua_remove_alias(OsipUA *ua, char *contact);

/* remove all aliases */
void osip_ua_clean_alias(OsipUA *ua);

/* set port */
void osip_ua_set_rtpport(OsipUA *ua,int type,int port);

/* clean profiles capabilities */
void osip_ua_clean_profiles(OsipUA *ua, int type);

/* add a profile to the ua capabilities */
void osip_ua_add_profile(OsipUA *ua,int type,int profile);

/* set a default registrar where REGISTER are sent. To unset, give registrar=NULL. */
int osip_ua_set_registrar( OsipUA *ua, char *registrar, char *passwd);

/* set a proxy to which all requests are sent. It does the same as osip_ua_set_registrar() but also indicate that the registrar
is also a proxy */
int osip_ua_set_outbound_proxy(OsipUA* ua,char *proxy,char *passwd);

/* set information about presence: this let the user agent to answer automatically
to incoming calls */
void osip_ua_set_presence_mode(OsipUA *ua, int mode);

/* set contact information used by presence modes 480, 301, 380.*/
void osip_ua_set_presence_contact_url(OsipUA *ua, char *url);

/* set the retry after */
void osip_ua_set_presence_delay(OsipUA *ua, int delay);

/*register callbacks to be advertised when an invite arrives, a sent invite is accepted, or a bye is received, or a faillure occured*/
int osip_ua_signal_connect(OsipUA *ua,char *signal,int (*func)(OsipCallLeg *,sip_t *msg, void*));

/* destroy and free all ressources allocated by user_agent*/
int osip_ua_destroy( OsipUA *ua);



#include "osipcallleg.h"

/**********************************************************************************************************************************************************************/
/*																				Private Functions: do not use them in an application																						*/
/***********************************************************************************************************************************************************************/

extern list_t ua_list;


/* find the user agent to which the message is destinated*/
OsipUA * osip_ua_find(sip_t *sipmsg);



#define AUDIO_PROFILES 0
#define VIDEO_PROFILES 1


#endif

