/*
  The oSIP library implements the Session Initiation Protocol (SIP -rfc2543-)
  Copyright (C) 2001  Aymeric MOIZARD jack@atosc.org

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <config.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include "utils.h"
#include <osip/sdp.h>
#include "osipmanager.h"
#include "resolver.h"
#include "uatransaction.h"




void respond_to_request(osip_t *config, transaction_t *transaction, int code, char *body)
{
	char *reason;
	char *apptype=NULL;
	sip_t *response;
	
	msg_init(&response);
	reason = msg_getreason(code);

	if (body!=NULL) apptype=sgetcopy("application/sdp");

    if (1==msg_makereply(code,reason,
			 transaction->lastrequest,
			 response,
			 body, /*local_body*/
			 apptype, /*apptype */
			 0))
	{
		sfree(reason);
		if (apptype!=NULL)
      		sfree(apptype);
    /* REVISIT here: we use def_manager instead of config */
		osip_send_response(def_manager,transaction,response);
	}
	else
	{
		osip_trace(TRACE_LEVEL1,("error: Could not build response for current status\n"));
		sfree(reason);
		if (apptype!=NULL)
      		sfree(apptype);
		sfree(response);
	}
	
}


void osip_send_response(OsipManager *mgr,transaction_t *transaction,sip_t *response)
{
	via_t *via;
	sipevent_t *sipevent;
	/* we must check if first via has x.x.x.x address. If not, we must resolve it*/
	msg_getvia(response,0,&via);
	if (via==NULL)
	{
		osip_trace(TRACE_LEVEL0,("error: Cannot send response because it has no via field"));
		return;
	}
		
	sipevent = osip_new_outgoing_sipmessage(response);
	sipevent->transactionid =  transaction->transactionid;

	if (inet_addr(via->host)==-1)
	{
		/* need name resolution*/
		async_resolv_and_send(def_manager,transaction,sipevent);
		return;
	}
	fifo_add(transaction->transactionff,sipevent);
}

char *sdp_compose(int sess_id, char *user, char *ip4addr,char *subject, list_t *media_list)
{
	char *sess;
	sdp_t *sdpmsg;
	timed_t *timed;
	mediad_t *media;
	MediaDesc *md;
	char *sdp;
	char *profs,*port;
	int i;
	
	
	sdpmsg=smalloc(sizeof(sdp_t));
	sdp_init(sdpmsg);
	/* v=0*/
	sdpmsg->v = sdp_initv("0");

	/*o= ...*/
	sess = (char *)smalloc(10);
	sprintf(sess,"%i",sess_id);
	sdpmsg->o = sdp_inito(user,
			sess,
			sess,
			"IN",
			"IP4",
			ip4addr);
	
	/*s= subject of converstation*/
	sdpmsg->s = sdp_inits(subject);

	/* c= ....*/	
	timed = (timed_t *)smalloc(sizeof(timed_t));
	timed_init(timed);
	timed->t = sdp_initt("0","0");
	list_add(sdpmsg->timeds,timed,-1);

	sdpmsg->c = sdp_initc("IN","IP4",ip4addr);
	/* m=...*/
	for (i=0;i<media_list->nb_elt;i++)
	{
		media=smalloc(sizeof(mediad_t));
		mediad_init(media);
		md=(MediaDesc*)list_get(media_list,i);
		profs=media_desc_profiles_2char(md);
		port=media_desc_port_2char(md);
		media->m = sdp_initm("audio",port,"RTP/AVP",profs);
		list_add(sdpmsg->mediads,media,-1);
	}
	/* convert to string */
	sdp=sdp_2char(sdpmsg);
	if (sdp==NULL) osip_trace(TRACE_LEVEL0,("error: converting sdpmsg to char.\n"));
	sdp_free(sdpmsg);
	sfree(sess);
	return(sdp);
}

/* fill default feilds for reply.                    */
/* OUTPUT: sip_t *dest | structure to store response.*/
/* INPUT : sip_t *request | previous request.        */
/* INPUT : int status | status for response.         */
/* INPUT : char *reason | reasonphrase for response. */
/* INPUT : char *body | body for response.           */
/* INPUT : int options | add a tag in to header.     */
/* return -1 on error */
/* TODO: add built-in support for record-route and route */
int
msg_makereply(int status,char *reason,sip_t *request,
	      sip_t *response,char *body,char *content_type
	      ,int options)
{
  int pos;
  int i;

  /* initialise sip_t structure */
  /* yet done... */

  response->strtline = (startline_t *)smalloc(sizeof(startline_t));
  response->strtline->sipversion = (char *)smalloc(8*sizeof(char));
  sprintf(response->strtline->sipversion,"SIP/2.0");
  response->strtline->statuscode = (char *)smalloc(5*sizeof(char));
  sprintf(response->strtline->statuscode,"%i",status);
  response->strtline->reasonphrase = (char *)smalloc(strlen(reason)+1);
  sprintf(response->strtline->reasonphrase,"%s",reason);

  response->strtline->rquri     = NULL;
  response->strtline->sipmethod = NULL;

  {
    char *tmp;
    i = to_2char(request->to, &tmp);
    if (i==-1) return -1;
    i = msg_setto(response, tmp);
    sfree(tmp);
    if (i==-1) return -1;

  }
  {
    char *tmp;
    i = from_2char(request->from, &tmp);
    if (i==-1) return -1;
    i = msg_setfrom(response, tmp);
    sfree(tmp);
    if (i==-1) return -1;
  }

  /* via headers */
  {
    char *tmp;
    pos = 0;
    while (!list_eol(request->vias,pos))
      {
	via_t *via;
	via = (via_t *)list_get(request->vias,pos);
	i = via_2char(via, &tmp);
	if (i==-1) return -1;
	i = msg_setvia(response, tmp);
	sfree(tmp);
	if (i==-1) return -1;
	pos++;
      }
  }

  {
    char *tmp;
    i = call_id_2char(request->call_id, &tmp);
    if (i==-1) return -1;
    i = msg_setcall_id(response, tmp);
    sfree(tmp);
    if (i==-1) return -1;
  }
  {
    char *tmp;
    i = cseq_2char(request->cseq, &tmp);
    if (i==-1) return -1;
    i = msg_setcseq(response, tmp);
    sfree(tmp);
    if (i==-1) return -1;
  }
  /* if a record-route header is found, set a route a header
     with the reverse order of urls  and copy the record route
     to the response. */
	{
		char *tmp;
		record_route_t *rr,*rr2;
		contact_t *ct=NULL;
		/* this test is useles!!! */
		/* if (request->record_routes!=NULL) */
		if (!list_eol(request->record_routes,0))
		{
			for(pos=request->record_routes->nb_elt-1;pos>=0;pos--)
			{
				rr=(record_route_t*)list_get(request->record_routes,pos);
				i=record_route_2char(rr,&tmp); 
				//printf("setting record route %s\n",tmp);
				/* BKO: do not set the route in responses */
/* 				i=msg_setroute(response,tmp); */
				record_route_init(&rr2);
				record_route_parse(rr2,tmp);
				list_add(response->record_routes,rr2,0);
				sfree(tmp);
			}
		/* if contact is set, we must add it to the Route header */
		/*
		*/
		msg_getcontact(request,0,&ct);
		if (ct!=NULL)
		{
			contact_2char(ct,&tmp);
			/* BKO: do not append the Contact: to route or
                           record-route in responses */
/*			msg_setroute(response,tmp); */
			/* NOT Route, but Record-Route */
		/*	msg_setrecord_route(response,tmp); */
			sfree(tmp);
		}
	}
}


	
    /* set user agent information*/
   msg_setuser_agent(response,"oSIP/Linphone-"VERSION);
   /* add personnal contact information (to be used for next request */
   /* (only for useragent) */
   if (body!=NULL)
     {
       char *size;
       i = msg_setbody(response, body);
       if (i==-1) return -1;

       size = (char *) smalloc(6*sizeof(char));
       sprintf(size,"%i",strlen(body));
       i = msg_setcontent_length(response, size);
       sfree(size);
       if (i==-1) return -1;

       {
	 char *ct = sgetcopy("Content-Type");
	 i = msg_setheader(response, ct, content_type);
	 sfree(ct);
	 if (i==-1) return -1;
       }
     }
   else
     {
       char *cl = sgetcopy("0");
       i = msg_setcontent_length(response, "0");
       sfree(cl);
       if (i==-1) return -1;
     }

   return 1;
}


int async_resolv_and_send(OsipManager *manager, transaction_t *transaction, sipevent_t *sipevent)
{
	 /* put the transaction,event on the fifo of the resolver thread */
	trans_event_t *te;
	
	osip_trace(TRACE_LEVEL4,("info: Name resolution requested.\n"));
	if (manager->resolv_run_cond==0) osip_manager_start_resolver(manager);
	te=smalloc(sizeof( trans_event_t));
	te->transaction=transaction;
	te->event=sipevent;
	fifo_add(manager->resolv_fifo,(void*)te);
	return transaction->transactionid;
}

/* profiles is a list of profiles supported by remote host, and ua_profile is the list for local user agent. The return value is the preferred profile for ua, compatible with
remote host 's list. */
int get_compatible_profile(list_t *ua_profiles,list_t *profiles)
{
	int i,j;
	int local_prof=0,remote_prof;
	int found=0;
	
	
	if (ua_profiles==NULL) return(-1);
	if (ua_profiles->nb_elt==0) return(-1);
	if (profiles==NULL) return(-1);
	if (profiles->nb_elt==0) return(-1);
	
	for(i=0;((i<ua_profiles->nb_elt) && (!found));i++)
	{
		local_prof=((int*)list_get(ua_profiles,i))[0];
		for (j=0;(j<profiles->nb_elt && !found) ;j++)
		{
			remote_prof=((int*)list_get(profiles,j))[0];
			if (remote_prof==local_prof)
			{
				found=1;
			}
		}
	}
	if (found) return(local_prof);
	else return(-1);
}
			
/*search element el in list, and remove it*/
int list_remove_el(list_t *list,void *el)
{
	int pos=0;
	int cond=0;
	node_t *nd;
	
	nd=list->node;
	while((pos<list->nb_elt)&&(!cond))
	{
		if (nd->element==el) cond=1;
		else
		{
			pos++;
			nd=nd->next;
		}
	}
	if (cond) list_remove(list,pos);
	else return(-1);
	return(0);
}
		
void callid_getcopy(call_id_t	*dest,call_id_t *source)
{
	dest->number=smalloc(strlen(source->number)+1);
	strcpy(dest->number,source->number);
	dest->host=smalloc(strlen(source->host)+1);
	strcpy(dest->host,source->host);
}

void
list_free(list_t *li, void (*destructor)(void *))
{
  int pos = 0;
  void *elem;
  while (!list_eol(li,pos))
    {
     elem = list_get(li,pos);
     list_remove(li,pos);
     if (destructor!=NULL) destructor(elem);
    }
}

void *
srealloc(void *p,int i)
{
  return realloc(p,i);
}

char *make_message(const char *fmt, ...)
{
	 /* Guess we need no more than 100 bytes. */
	int n, size = 100;
	char *p;
	va_list ap;
	if ((p = smalloc (size)) == NULL)
		return NULL;
	while (1) {
	/* Try to print in the allocated space. */
	va_start(ap, fmt);
	n = vsnprintf (p, size, fmt, ap);
	va_end(ap);
	/* If that worked, return the string. */
	if (n > -1 && n < size)
		return p;
	//printf("Reallocing space.\n");
	/* Else try again with more space. */
	if (n > -1)    /* glibc 2.1 */
		size = n+1; /* precisely what is needed */
	else           /* glibc 2.0 */
		size *= 2;  /* twice the old size */
	if ((p = srealloc (p, size)) == NULL)
		return NULL;
	}
}

/* from2 has tag parameter that we should copy to from1 if it does not have one */
int from_copy_tag(from_t *from1,from_t *from2)
{
	generic_param_t *param1=NULL,*param2=NULL;
	from_param_getbyname(from1,"tag",&param1);
	if (param1==NULL)
	{
		from_param_getbyname(from2,"tag",&param2);
		if (param2!=NULL)
		{
			//printf("getting from tag field!!!\n");
			from_param_add(from1,sgetcopy("tag"),sgetcopy(param2->gvalue));
		}
	}
	return 0;
}

int _tag_init(char **tag,int length)
{
	char *newtag=smalloc((length+2)*sizeof(char));
	snprintf(newtag,length+1,"%x",rand());
	*tag=newtag;
	return 0;
}

/* this function check is *url has a sip: prefix. If it has not, then it add ones in a newly allocated string. If no correction is made, it just returns
the initial string */
char * check_sipurl(char *url)
{
	char *newstr;
	from_t *from;
	int err;
	
	/* try to parse to see if the address in correct */
	from_init(&from);
	err=from_parse(from,url);
	from_free(from);
	sfree(from);	
	if (err==0)	return url;  /* OK */
	
	
	if (strncmp(url,"sip:",4)==0) return NULL;
	if (strncmp(url,"<sip:",5)==0) return NULL;

   newstr=smalloc(strlen(url)+5);
   strcpy(newstr,url);
   sclrspace(newstr);
   if (strncmp(newstr,"sip:",4)==0) return newstr;
	if (strncmp(newstr,"<sip:",5)==0) return newstr;
	/* if we are here, there is no "sip:" at the begginning, so add it */
	
	if (newstr[0]=='<')  sprintf(newstr,"<sip:%s",url+1);   /* do a <sip:machin@truc> */
	else  sprintf(newstr,"sip:%s",url);   /* do a sip:machin@truc */
	sclrspace(newstr);
	/* try to parse to see if the address in correct */
	from_init(&from);
	err=from_parse(from,newstr);
	from_free(from);
	sfree(from);	
	if (err==0)	return newstr;  /* OK */
	else return (NULL);
}






